module Hawl.Util.CGI where

import Hawl.Util
import Hawl.Util.List
import Hawl.Util.XHtml

import Control.Monad (liftM)
import Control.Monad.Trans (MonadIO)
import Control.Exception (Exception)
import Data.List
import Data.Maybe (fromMaybe)
import Network.URI (escapeURIString, isUnescapedInURI)

import Data.ByteString.Lazy (ByteString)

import Network.NewCGI hiding (Html)
import Text.XHtml

--
-- * Request method
--

type MethodHandler m a = (String, m a)

getMethod :: MonadCGI m => m String 
getMethod = liftM (fromMaybe "GET") (getVar "REQUEST_METHOD")

methodHandlers :: (MonadIO m, MonadCGI m) => 
                  (a -> m CGIResult) -> [MethodHandler m a] -> m CGIResult
methodHandlers f hs = 
    do m <- getMethod
       case lookup m hs of
           Just h  -> h >>= f
           Nothing -> errorMethodNotAllowed (map fst hs)

get :: m a -> MethodHandler m a
get x = ("GET", x)

post :: m a -> MethodHandler m a
post x = ("POST", x)

--
-- * Inputs
--

getInputWithDefault :: MonadCGI m => 
                       String -- ^ Default value
                    -> String -> m String
getInputWithDefault d = liftM (fromMaybe d) . getInput

getInputOrFail :: MonadCGI m => String -> m String
getInputOrFail n = getInput n >>= maybeM ("Missing input: " ++ n)

getInputFPSOrFail :: MonadCGI m => String -> m ByteString
getInputFPSOrFail n = getInputFPS n >>= maybeM ("Missing input: " ++ n)

getManyInputs :: MonadCGI m => [String] -> m [String]
getManyInputs = mapM getInputOrFail

readInputWithDefault :: (MonadCGI m, Read a) => a -> String -> m a
readInputWithDefault d = liftM (fromMaybe d) . readInput

readMultiInput :: (MonadCGI m, Read a) => String -> m [a]
readMultiInput n = getMultiInput n >>= mapM readM

--
-- * Output
--

outputHtml :: MonadCGI m => Html -> m CGIResult
outputHtml = output . renderHtml

--
-- * Error handling
--

cgiFail :: (MonadCGI m,MonadIO m) => Int -> String -> [String] -> m CGIResult
cgiFail c s ss = 
      do setHeader "Status" (show c)
         logCGI (show ss)
         let page = errorPage (show c ++ " " ++ s) ss
         output $ renderHtml page 

cgiErrorHandler :: (MonadCGI m,MonadIO m) => Exception -> m CGIResult
cgiErrorHandler e = errorServer (show e)

errorPage :: String -> [String] -> Html
errorPage tit ps = 
    mkHtml (thetitle << tit) ((h1 << tit) : map (paragraph <<) ps)

runCGIWithErrorHandler :: CGI CGIResult -> IO ()
runCGIWithErrorHandler f = runCGI (f `handleExceptionCGI` cgiErrorHandler)

--
-- * Specific HTTP errors
--

errorNotFound :: (MonadIO m, MonadCGI m) => String -> m CGIResult
errorNotFound r =
    cgiFail 404 "Not Found" ["The requested resource was not found: " ++ r]

errorMethodNotAllowed :: (MonadIO m, MonadCGI m) => [String] -> m CGIResult
errorMethodNotAllowed ms = 
    do setHeader "Allow" $ glue " ," ms
       cgiFail 405 "Method Not Allowed" []

errorServer :: (MonadIO m, MonadCGI m) => String -> m CGIResult
errorServer e = cgiFail 500 "Internal Server Error" [e]

--
-- * URL encoding
--

-- | Format name-value pairs as application\/x-www-form-urlencoded.
formEncode :: [(String,String)] -> String
formEncode xs = 
    concat $ intersperse "&" [urlEncode n ++ "=" ++ urlEncode v | (n,v) <- xs]

-- | Convert a single value to the application\/x-www-form-urlencoded encoding.
urlEncode :: String -> String
urlEncode = replace ' ' '+' . escapeURIString okChar
  where okChar c = c == ' ' || 
                   (isUnescapedInURI c && c `notElem` "&=+")
